<?php

namespace InSegment\ApiCore\Api;

use InSegment\ApiCore\Facades\DTODefs;

/**
 * Govern recursive actions on rules
 */
abstract class RuleAnalysis
{
    /**
     * A collection of conversion rules by class
     * 
     * @var array
     */
    protected $rules;
    
    /**
     * DTODefs instance. To be able to pass by reference
     * 
     * @var \InSegment\ApiCore\Facades\DTODefs
     */
    protected $dtoDefs;
    
    /**
     * Descending param names (in straight order) to their variants with capitalized first letter
     * 
     * @var array 
     */
    private $descendingParamNames = [];

    /**
     * Ascending param names (in reverse order) to their variants with capitalized first letter
     * 
     * @var array 
     */
    private $ascendingParamNames;
    
    /**
     * Constructor
     */
    public function __construct(&$rules)
    {
        $this->rules = &$rules;
        $this->dtoDefs = DTODefs::instance();
        
        foreach ($this->getStackingParamNames() as $param) {
            $this->descendingParamNames[$param] = ucfirst($param);
        }
        
        $this->ascendingParamNames = array_reverse($this->descendingParamNames);
    }
    
    /**
     * Get rules for class
     * 
     * @param string $class
     * @return array
     */
    public function &getRules($class)
    {
        return $this->rules[$class];
    }
    
    /**
     * Should return an array of parameters acceptable for descend method of the concrete implementation
     * For each such method the concrete implementation must have methods:
     * 
     * function descendParamName($ascendedPreviousValue)
     * function ascendParamName($paramValue)
     * 
     * @return array
     */
    protected abstract function getStackingParamNames(): array;
    
    /**
     * Do an action descending by dotFormat into input
     * 
     * @param array $params
     * @param callable $closure
     * @param array $parameters
     * @return mixed
     */
    protected function descend(array $params, callable $closure, ...$parameters)
    {
        $ascended = [];
        
        // Descend everything
        foreach ($this->descendingParamNames as $param => $ucParam) {
            if (isset($params[$param])) {
                $ascended[$ucParam] = $this->{"descend{$ucParam}"}($params[$param]);
            }
        }
        
        // run action
        //file_put_contents(base_path().'/stacklog2.txt', date('D M j H:i:s Y Descend:') . print_r([$dotFormat ?? null], true), FILE_APPEND);
        $result = $closure(...$parameters);
        //file_put_contents(base_path().'/stacklog2.txt', date('D M j H:i:s Y Ascend:') . print_r([$dotFormat ?? null], true), FILE_APPEND);
        
        // Ascend everything
        foreach ($this->ascendingParamNames as $param => $ucParam) {
            if (isset($params[$param])) {
                $this->{"ascend{$ucParam}"}($ascended[$ucParam]);
            }
        }
        
        return $result;
    }
    
    /**
     * Return array consisting of properties of $this
     * 
     * @param array $props
     * @return array
     */
    protected function compactProps(...$props): array
    {
        return array_reduce($props, function ($carry, $prop) {
            $carry[$prop] = $this->$prop;
            return $carry;
        }, []);
    }
    
    /**
     * Set properties of $this by array
     * 
     * @param array $compactedAscneded
     * @return null
     */
    protected function extractProps(array &$compactedAscneded)
    {
        foreach ($compactedAscneded as $prop => &$value) {
            $this->$prop = &$value;
        }
    }
}
